#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <maibu_sdk.h>
#include "maibu_res.h"

typedef struct Rational {
    int num;
    int den;
}Rational;

typedef uint8_t Index;

typedef struct Solution {
    Index arrange;
    Index operator;
    Index bracket;
}Solution;


#define NUMBER_COUNT 4

//占用24字节
#define ARRANGE_COUNT 24
#define ARRANGE_SIZE NUMBER_COUNT
#define A(a,b,c,d) ((a<<6)|(b<<4)|(c<<2)|(d))
typedef uint8_t CompressArrange;
static CompressArrange arranges[ARRANGE_COUNT];
//  = {
//     A(0,1,2,3),A(0,1,3,2),A(0,2,1,3),A(0,2,3,1),
//     A(1,0,2,3),A(1,0,3,2),A(1,2,0,3),A(1,2,3,0),
//     A(0,3,1,2),A(0,3,2,1),A(1,3,0,2),A(1,3,2,0),
//     A(2,0,1,3),A(2,0,3,1),A(2,1,0,3),A(2,1,3,0),
//     A(2,3,0,1),A(2,3,1,0),A(3,0,1,2),A(3,0,2,1),
//     A(3,1,0,2),A(3,1,2,0),A(3,2,0,1),A(3,2,1,0)
// };

//占用4+4*4+64=84字节
#define OPERATOR_COUNT 64
#define OPERATOR_SIZE 3
typedef Rational (*Operator)(Rational a1, Rational a2);
typedef uint8_t CompressOperator;//00 01 10 11(第一个两位不使用，后面每两位表示一个Index)
static char operator_names[4];// = {'+', '-', '*', '/'};
static Operator operators[4];// = {op_add, op_minus, op_mult, op_div};
#define B(b,c,d) ((b<<4)|(c<<2)|(d))
static CompressOperator cops[OPERATOR_COUNT];
//  = {
//     B(0,0,0),B(1,1,1),B(2,2,2),B(3,3,3),

//     B(0,0,1),B(0,1,0),B(1,0,0),
//     B(0,0,2),B(0,2,0),B(2,0,0),
//     B(0,0,3),B(0,3,0),B(3,0,0),

//     B(1,1,0),B(1,0,1),B(0,1,1),
//     B(1,1,2),B(1,2,1),B(2,1,1),
//     B(1,1,3),B(1,3,1),B(3,1,1),

//     B(2,2,0),B(2,0,2),B(0,2,2),
//     B(2,2,1),B(2,1,2),B(1,2,2),
//     B(2,2,3),B(2,3,2),B(3,2,2),

//     B(3,3,0),B(3,0,3),B(0,3,3),
//     B(3,3,1),B(3,1,3),B(1,3,3),
//     B(3,3,2),B(3,2,3),B(2,3,3),

//     B(0,1,2),B(0,2,1),B(1,0,2),B(1,2,0),B(2,0,1),B(2,1,0),
//     B(0,1,3),B(0,3,1),B(1,0,3),B(1,3,0),B(3,0,1),B(3,1,0),
//     B(0,2,3),B(0,2,3),B(3,0,2),B(3,2,0),B(2,0,3),B(2,3,0),
//     B(1,2,3),B(3,2,1),B(1,3,2),B(1,2,3),B(2,3,1),B(2,1,3),
// };

//占用5*8=40字节
#define BRACKET_COUNT 5
#define MAX_BUF_SIZE 20
typedef Rational (*Bracket)(Rational nums[NUMBER_COUNT], Operator ops[OPERATOR_SIZE]);
//A(B(CD)), A((BC)D), (AB)(CD), (A(BC))D, ((AB)C)D
static const char* brakcet_fmts[BRACKET_COUNT];
// = {
    // "%d%c(%d%c(%d%c%d))", 
    // "%d%c((%d%c%d)%c%d)",
    // "(%d%c%d)%c(%d%c%d)",
    // "(%d%c(%d%c%d))%c%d",
    // "((%d%c%d)%c%d)%c%d"
    // };
static Bracket brackets[BRACKET_COUNT];// = {bk1, bk2, bk3, bk4, bk5};

static void get_ops(CompressOperator c, Operator ops[OPERATOR_SIZE]) {
    ops[0] = operators[c&0x3];
    ops[1] = operators[(c&0x0C)>>2];
    ops[2] = operators[(c&0x30)>>4];
}

static void get_ops_name(CompressOperator c, char names[OPERATOR_SIZE]) {
    names[0] = operator_names[c&0x3];
    names[1] = operator_names[(c&0x0C)>>2];
    names[2] = operator_names[(c&0x30)>>4];
}

static void get_arrange_as_rational(int n[NUMBER_COUNT], Rational r[NUMBER_COUNT], Index arrange) {
    CompressArrange a = arranges[arrange];
    int index[NUMBER_COUNT];
    Index i = 0;

    index[0] = a&0x3;
    index[1] = (a&0x0C)>>2;
    index[2] = (a&0x30)>>4;
    index[3] = (a&0xC0)>>6;
    
    for (; i < NUMBER_COUNT; i++) {
        r[i].den = 1;
        r[i].num = n[index[i]];
    }
}

static int is_24(Rational r) {
    if (r.den == 0)
        return 0;
    return r.num / r.den == 24 && r.num % r.den == 0;
}

static void print_solution(int number[NUMBER_COUNT], Solution* solution, char buf[MAX_BUF_SIZE]) {
    Rational nums[NUMBER_COUNT];
    get_arrange_as_rational(number, nums, solution->arrange);

    char ops[OPERATOR_SIZE];
    get_ops_name(cops[solution->operator], ops);

    sprintf(buf, brakcet_fmts[solution->bracket], nums[0].num, ops[0], nums[1].num, ops[1], nums[2].num, ops[2], nums[3].num);
}

static int solve(int number[NUMBER_COUNT], Solution* solution) {
    Index i = 0;
    for (; i < ARRANGE_COUNT; i++){//遍历可能的排列
        Index j = 0;
        Rational nums[NUMBER_COUNT];
        get_arrange_as_rational(number, nums, i);

        for (; j < OPERATOR_COUNT; j++) {//遍历可能的操作符组合
            Index k = 0;
            Operator ops[OPERATOR_SIZE];
            get_ops(cops[j], ops);

            for (; k < BRACKET_COUNT; k++) {//遍历所有可能的括号方法
                Bracket bracket = brackets[k];
                Rational ret = bracket(nums, ops);
                if (!is_24(ret))
                    continue;

                solution->arrange = i;
                solution->operator = j;
                solution->bracket = k;
                return 1;
            }
        }
    }
    return 0;
}

static Rational op_add(Rational a1, Rational a2){
    Rational r;
    r.den = a1.den * a2.den;
    r.num = a1.num * a2.den + a2.num * a1.den;
    return r;
}
static Rational op_minus(Rational a1, Rational a2){
    Rational r;
    r.den = a1.den * a2.den;
    r.num = a1.num * a2.den - a2.num * a1.den;
    return r;
}
static Rational op_mult(Rational a1, Rational a2){
    Rational r;
    r.den = a1.den * a2.den;
    r.num = a1.num * a2.num;
    return r;
}
static Rational op_div(Rational a1, Rational a2){
    Rational r;
    r.den = a1.den * a2.num;
    r.num = a1.num * a2.den;
    return r;
}

static Rational bk1(Rational nums[NUMBER_COUNT], Operator ops[OPERATOR_SIZE]) {
    //A(B(CD))
    Rational cd = ops[2](nums[2], nums[3]);
    Rational bcd = ops[1](nums[1], cd);
    return ops[0](nums[0], bcd);
}
static Rational bk2(Rational nums[NUMBER_COUNT], Operator ops[OPERATOR_SIZE]) {
    //A((BC)D)
    Rational bc = ops[1](nums[1],nums[2]);
    Rational bcd = ops[2](bc, nums[3]);
    return ops[0](nums[0], bcd);
}
static Rational bk3(Rational nums[NUMBER_COUNT], Operator ops[OPERATOR_SIZE]) {
    //(AB)(CD)
    Rational ab = ops[0](nums[0], nums[1]);
    Rational cd = ops[2](nums[2], nums[3]);
    return ops[1](ab, cd);
}
static Rational bk4(Rational nums[NUMBER_COUNT], Operator ops[OPERATOR_SIZE]) {
    //(A(BC))D
    Rational bc = ops[1](nums[1],nums[2]);
    Rational abc = ops[0](nums[0], bc);
    return ops[2](abc, nums[3]);
}
static Rational bk5(Rational nums[NUMBER_COUNT], Operator ops[OPERATOR_SIZE]) {
    //((AB)C)D
    Rational ab = ops[0](nums[0], nums[1]);
    Rational abc = ops[1](ab, nums[2]);
    return ops[2](abc, nums[3]);
}

static void init_arrange() {
    // FILE* fp = fopen("arrange.dat", "rb");
    // fread(arranges, sizeof(CompressArrange), sizeof(arranges), fp);
    // fclose(fp);

    // printf("read arrange: %u\n", maibu_read_user_file(FILE_ARRANGE, 0, arranges, sizeof(CompressArrange) * sizeof(arranges)));
    maibu_read_user_file(FILE_ARRANGE, 0, arranges, sizeof(CompressArrange) * sizeof(arranges));
}

static void init_operation() {
    memcpy(operator_names, "+-*/", 4);
    operators[0] = op_add;
    operators[1] = op_minus;
    operators[2] = op_mult;
    operators[3] = op_div;

    // FILE* fp = fopen("operation.dat", "rb");
    // fread(cops, sizeof(CompressOperator), sizeof(cops), fp);
    // fclose(fp);

    // printf("read operators:%u\n", maibu_read_user_file(FILE_OPERATOR, 0, cops, sizeof(CompressOperator) * sizeof(cops)));
    maibu_read_user_file(FILE_OPERATOR, 0, cops, sizeof(CompressOperator) * sizeof(cops));
}

static void init_bracket() {
    brakcet_fmts[0] = "%d%c(%d%c(%d%c%d))";
    brakcet_fmts[1] = "%d%c((%d%c%d)%c%d)";
    brakcet_fmts[2] = "(%d%c%d)%c(%d%c%d)";
    brakcet_fmts[3] = "(%d%c(%d%c%d))%c%d";
    brakcet_fmts[4] = "((%d%c%d)%c%d)%c%d";

    brackets[0] = bk1;
    brackets[1] = bk2;
    brackets[2] = bk3;
    brackets[3] = bk4;
    brackets[4] = bk5;
}

static void init_magic() {
    init_arrange();
    init_operation();
    init_bracket();    
}

// int main(int argc, char* argv[]) {
//     init_magic();

//     int number[NUMBER_COUNT] = {8, 3, 7, 1};
//     Solution solution;

//     if (solve(number, &solution)){
//         char buf[MAX_BUF_SIZE]={0};
//         print_solution(number, &solution, buf);
//         printf("got a solution: %s\n", buf);
//     }
//     else{
//         printf("no solution found\n");
//     }

//     // FILE* fp = fopen("arrange.dat", "wb");
//     // fwrite(arranges, sizeof(CompressArrange), sizeof(arranges), fp);
//     // fclose(fp);

//     // fp = fopen("operation.dat", "wb");
//     // fwrite(cops, sizeof(CompressOperator), sizeof(cops), fp);
//     // fclose(fp);

//     return 0;
// }