#include <maibu_sdk.h>
#include <stdio.h>

// #define SIMULATE
#include "watch_def.h"
#include "build/ui.h"
#include "game.h"

static char g_answer[MAX_BUF_SIZE];

typedef struct _Card {
    uint8_t bmpId;
    uint8_t textId;
}Card;

typedef struct _Anim {
    uint8_t timerid;
    Index index;
}Anim;

static Card g_cards[NUMBER_COUNT];
static Anim g_anim;
static uint8_t g_number[NUMBER_COUNT];

static void open_card(P_Window window, Card* card, int num) {
    LAYER_SET_BMP(card->bmpId, BMP_FRONT);

    char num_str[3] = {0};
    if (num == 1) 
        num_str[0] = 'A';
    else if (num == 11) 
        num_str[0] = 'J';
    else if (num == 12) 
        num_str[0] = 'Q';
    else if (num == 13) 
        num_str[0] = 'K';
    else
        sprintf(num_str, "%d", num);

    LAYER_SET_TEXT(card->textId, num_str);
    maibu_layer_set_visible_status(app_window_get_layer_by_id(window, card->textId), true);
}

static void close_card(P_Window window, Card* card) {
    maibu_layer_set_visible_status(app_window_get_layer_by_id(window, card->textId), false);
    LAYER_SET_BMP(card->bmpId, BMP_BACK);
}

void anim_callback(date_time_t dt, uint32_t millis ,void* context) {
    if (g_anim.index >= NUMBER_COUNT) {
        return;
    }

    P_Window window = (P_Window)context;
    open_card(window, &(g_cards[g_anim.index]), g_number[g_anim.index]);
    app_window_update(window);
    g_anim.index++;
}

static void start_anim() {
    g_anim.index = 0;
}

static void on_key_back(void* context) {
    Window* window = (Window*)context;
    if (window == NULL)
        return;
	
    app_window_stack_pop(window);
}

static void show_answer(P_Window window) {
    LAYER_SET_TEXT(g_hint_textid, g_answer);
}

static void on_key_ok(void* context) {
    Window* window = (Window*)context;
    if (window == NULL)
        return;
    
    show_answer(window);
    app_window_update(window);
}

static uint32_t g_next = 0;
static int my_rand() {
    g_next = g_next * 1103515245+12345;
    return (unsigned int)(g_next/65536)%32768;
}
static int my_srand(unsigned int seed) {
    g_next = seed;
}

static void next_puzzle(P_Window window) {
    int number[NUMBER_COUNT] = {0};
    Solution solution;
    int i = 0;

    for (;i<NUMBER_COUNT;i++) {
        g_number[i] = number[i] = my_rand()%13+1;
        close_card(window, &(g_cards[i]));
    }

    if (solve(number, &solution)){
        char buf[MAX_BUF_SIZE]={0};
        print_solution(number, &solution, g_answer);
        LAYER_SET_TEXT(g_hint_textid, "*********");
        start_anim();
    }
    else{
        for (i=0; i < NUMBER_COUNT; i++)
            open_card(window, &(g_cards[i]), number[i]);
        LAYER_SET_TEXT(g_hint_textid, "无解");
    }
    
}

static void on_key_down(void* context) {
    Window* window = (Window*)context;
    if (window == NULL)
        return;
    
    next_puzzle(window);
    app_window_update(window);
}

/**
 * 优化随机数
 * 翻牌动画
 */
void watch_main(){
    init_magic();

    P_Window window = create_main_window();
    Index i = 0;

    g_cards[0].bmpId = g_card1_bmpid;
    g_cards[0].textId = g_text1_textid;
    g_cards[1].bmpId = g_card2_bmpid;
    g_cards[1].textId = g_text2_textid;
    g_cards[2].bmpId = g_card3_bmpid;
    g_cards[2].textId = g_text3_textid;
    g_cards[3].bmpId = g_card4_bmpid;
    g_cards[3].textId = g_text4_textid;

    for (; i < NUMBER_COUNT; i++){
        close_card(window, &(g_cards[i]));
    }
    g_anim.index = NUMBER_COUNT;

    app_window_click_subscribe(window, ButtonIdBack, on_key_back);
    app_window_click_subscribe(window, ButtonIdDown, on_key_down);
    app_window_click_subscribe(window, ButtonIdSelect, on_key_ok);

    struct date_time dt;
    app_service_get_datetime(&dt);
    my_srand(dt.hour * 3600 + dt.min * 60 + dt.sec);

    app_window_stack_push(window);

    g_anim.timerid = app_service_timer_subscribe(200, anim_callback, window);
}